/* General system functions
   Copyright 2003-2005	Craig Franklin
   Copyright 2015-2016	Molnár Károly
*/

#include "stdhdr.h"
#include "libgputils.h"

#ifdef HAVE_WINDOWS_H
#include <windows.h>
#endif

#ifdef STDC_HEADERS
#include <stdarg.h>
#endif

const char *gp_header_path;
const char *gp_lkr_path;
const char *gp_lib_path;

bool absolute_path_warning = true;

/*------------------------------------------------------------------------------------------------*/

/* initialize the library */

FUNC(void) gp_init(void) {
  /* load the environmental variables */
  gp_header_path = getenv("GPUTILS_HEADER_PATH");
  gp_lkr_path    = getenv("GPUTILS_LKR_PATH");
  gp_lib_path    = getenv("GPUTILS_LIB_PATH");
}

/*------------------------------------------------------------------------------------------------*/

/* little endian functions */

void gp_fputl16(int Data, FILE *Fp) {
  assert(-32678<=Data && Data<32768);
  fwrite(&Data,1,2,Fp);
}

/*------------------------------------------------------------------------------------------------*/

void gp_fputl32(int Data, FILE *Fp) {
  fwrite(&Data,1,4,Fp);
}

/*------------------------------------------------------------------------------------------------*/

void gp_fputzero(size_t Number, FILE *Fp) {
  while (Number > 0) {
    fputc(0, Fp);
    --Number;
  }
}

/*------------------------------------------------------------------------------------------------*/

void gp_fputvar(const void *Data, size_t Number, FILE *Fp) {
  const uint8_t *d;
  size_t         i;

  d = (const uint8_t *)Data;
  for (i = 0; i < Number; i++) {
    fputc(d[i], Fp);
  }
}

/* big endian functions */

FUNC(int32_t) gp_getb32(const uint8_t *Addr) {
  int32_t value;

  value  = (int32_t)Addr[0] << 24;
  value |= (int32_t)Addr[1] << 16;
  value |= (int32_t)Addr[2] <<  8;
  value |= (int32_t)Addr[3];

  return value;
}

/*------------------------------------------------------------------------------------------------*/
// needed by gpcod.c
void gp_putb32(uint8_t *Addr, uint32_t Data) {
  Addr[0] = (Data >> 24) & 0xff;
  Addr[1] = (Data >> 16) & 0xff;
  Addr[2] = (Data >>  8) & 0xff;
  Addr[3] = Data & 0xff;
}

/*------------------------------------------------------------------------------------------------*/

FUNC(bool) gp_num_range_is_overlapped(int Area_start, int Area_end, int Ref_start, int Ref_end) {
  int min;
  int max;
  int w0;
  int w1;
  int width0;
  int width1;

  if (Area_start > Area_end) {
    /* Swap these values. */
    int w0     = Area_start;
    Area_start = Area_end;
    Area_end   = w0;
  }

  if (Ref_start > Ref_end) {
    /* Swap these values. */
    int w0    = Ref_start;
    Ref_start = Ref_end;
    Ref_end   = w0;
  }

  min = (Area_start < Ref_start) ? Area_start : Ref_start;
  max = (Area_end > Ref_end) ? Area_end : Ref_end;

  /*
     min  Area                         max
      |    |                            |
      V    V                  Ref       |
      +----------+             |        |
           w0                  V        V
                      +-----------------+
                              w1

      +---------------------------------+
                    width0
  */

  w0 = Area_end - Area_start + 1;
  w1 = Ref_end - Ref_start + 1;

  width0 = max - min + 1;
  width1 = w0 + w1;

  return width0 < width1;
}

/*------------------------------------------------------------------------------------------------*/

//unsigned gp_find_lowest_bit(unsigned Bits) {
//  unsigned i;

//  if (!Bits) return sizeof(unsigned)<<3;
//  for (i=0; !(Bits&1); Bits>>=1,i++);	// bit-scan-forward or -reverse?
//  return i;
//}

/*------------------------------------------------------------------------------------------------*/

//unsigned gp_find_highest_bit(unsigned Bits) {
//  unsigned i = sizeof(unsigned)<<3;
//  const unsigned msb = 1U<<(i-1);

//  if (!Bits) return i;
//  for (;--i,!(Bits&msb);Bits<<=1);
//  return i;
//}

/*------------------------------------------------------------------------------------------------*/

FUNC(void) gp_date_string(char *Buffer, size_t Sizeof_buffer) {
  time_t           now;
  const struct tm *now_tm;
#if defined(HAVE_LOCALE_H) && defined(HAVE_LANGINFO_H)
  char             format[60];
#endif

  time(&now);
  now_tm = localtime(&now);
  if (now_tm) {
#if defined(HAVE_LOCALE_H) && defined(HAVE_LANGINFO_H)
    setlocale(LC_ALL, "");
    snprintf(format, sizeof(format), "%s  %s", nl_langinfo(D_FMT), nl_langinfo(T_FMT));
    strftime(Buffer, Sizeof_buffer, format, now_tm);
    setlocale(LC_ALL, "C");
#else
    snprintf(Buffer, Sizeof_buffer,
             "%d-%d-%d  %02d:%02d:%02d",
             now_tm->tm_mon + 1,
             now_tm->tm_mday,
             1900 + now_tm->tm_year,
             now_tm->tm_hour,
             now_tm->tm_min,
             now_tm->tm_sec);
#endif
  }
  else if (Sizeof_buffer > 0) {
    Buffer[0] = '\0';
  }
}

/*------------------------------------------------------------------------------------------------*/

FUNC(void*) gp_malloc(size_t Size, const char *File, size_t Line) {
  void *m;

  if (!Size) return NULL;

  if ((m = malloc(Size)) == NULL) {
    fprintf(stderr, "%s.%s(%u) -- Could not allocate %u bytes of memory. {%s:%u}\n",
            __FILE__, "malloc", Size, File, Line);
    exit(1);
  }

  return m;
}

/*------------------------------------------------------------------------------------------------*/
/*
FUNC(void*) gp_calloc(size_t Nmemb, size_t Size, const char *File, size_t Line) {
  void *m;

  if (!Nmemb || !Size) return NULL;

  if ((m = calloc(Nmemb, Size)) == NULL) {
    fprintf(stderr, "%s.%s() -- Could not allocate %" SIZE_FMTu " bytes of memory. {%s.LINE-%" SIZE_FMTu ", %s()}\n",
            __FILE__, __func__, Nmemb * Size, File, Line, Func);
    exit(1);
  }

  return m;
}
*/
/*------------------------------------------------------------------------------------------------*/

FUNC(void*) gp_realloc(void *Mem, size_t Size, const char *File, size_t Line) {
  void *m;

  if (!Size) {free(Mem); return 0;}

  if (!(m = realloc(Mem, Size))) {
    fprintf(stderr, "%s.%s(%p,%u) -- Could not reallocate %u bytes of memory. {%s:%u}\n",
            __FILE__, "realloc", Mem, Size, Size, File, Line);
    exit(1);
  }

  return m;
}

/*------------------------------------------------------------------------------------------------*/

FUNC(char*) gp_strdup(const char *String, const char *File, size_t Line) {
  char *s;

  if (!String) return NULL;

  if ((s = strdup(String)) == NULL) {
    fprintf(stderr, "%s.%s(\"%s\") -- Could not allocate string {%s:%u}, error: %s.\n",
            __FILE__, "strdup", String, File, Line, strerror(errno));
    exit(1);
  }

  return s;
}

/*------------------------------------------------------------------------------------------------*/

FUNC(char*) gp_strndup(const char *String, size_t Length, const char *File, size_t Line) {
  char *s;

  if (!String) return NULL;

  if ((s = strndup(String, Length)) == NULL) {
    fprintf(stderr, "%s:%s(\"%s\",%u) -- Could not allocate string {%s:%u}, error: %s.\n",
            __FILE__, "strndup", String, Length, File, Line, strerror(errno));
    exit(1);
  }

  return s;
}

/*------------------------------------------------------------------------------------------------*/

FUNC(char*) gp_strdup_lower_case(const char *Name) {
  char  ch;
  char *pnew;
  char *ptr;

  ptr = pnew = GP_Strdup(Name);

  if (ptr) while ((ch = *ptr) != '\0') {
    *ptr = (tolower)(ch);
    ptr++;
  }

  return pnew;
}

/*------------------------------------------------------------------------------------------------*/

FUNC(char*) gp_strdup_upper_case(const char *Name) {
  char  ch;
  char *pnew;
  char *ptr;

  ptr = pnew = GP_Strdup(Name);

  if (ptr) while ((ch = *ptr) != '\0') {
    *ptr = (toupper)(ch);
    ptr++;
  }

  return pnew;
}

/*------------------------------------------------------------------------------------------------*/

FUNC(char*) gp_strncpy(char *Dest, const char *Src, size_t Maxlen) {
  char ch;

  assert(Dest);
  assert(Src);

  if (!Maxlen) return NULL;

  do {
    if (--Maxlen == 0) {
      *Dest = '\0';
      break;
    }

    ch = *Src++;
    *Dest++ = ch;
  } while (ch != '\0');

  return Dest;
}

/*------------------------------------------------------------------------------------------------*/

FUNC(char*) gp_arch_strncpy(char *Dest, const char *Src, size_t Maxlen) {
  char ch;

  assert(Dest);
  assert(Src);

  if (!Maxlen) return NULL;

  while (true) {
    if (--Maxlen == 0) {
      break;
    }

    ch = *Src;
    if (ch == '\0') {
      break;
    }

    *Dest = ch;
    ++Src;
    ++Dest;
  };

  return Dest;
}

/*------------------------------------------------------------------------------------------------*/

FUNC(char*) gp_stptoupper(char *Dest, const char *Src, size_t Maxlen) {
  char ch;

  assert(Dest);
  assert(Src);

  if (!Maxlen) return NULL;

  do {
    if (--Maxlen == 0) {
      *Dest = '\0';
      break;
    }

    ch = *Src++;
    *Dest++ = (toupper)(ch);
  } while (ch != '\0');

  return Dest;
}

/*------------------------------------------------------------------------------------------------*/

/*
    C_str            : Beginning address of C style string area.
    C_max_size       : Size of C style string area.
    Pascal_str       : Beginning address of Pascal style string area.
    Pascal_max_size  : Size of Pascal style string area, included the string length.
    Is_limited_length: This a pointer wherewith the function indicates that limitation occurred.
*/

FUNC(char*) gp_str_from_Pstr(char *C_str, size_t C_max_size, const uint8_t *Pascal_str, size_t Pascal_max_size,
                 bool *Is_limited_length) {
  size_t     length;
  bool limit;

  assert(C_str);
  assert(Pascal_str);

  /* This is a Pascal style string, read the length. */
  length = *Pascal_str++;	  /* The real beginning of string. */
  limit = false;

  if (length >= Pascal_max_size) {
    /* The string too long, corrects the size. */
    length = Pascal_max_size - 1;
    limit  = true;
  }

  if (length >= C_max_size) {
    /* The storage area too little. */
    length = C_max_size - 1;
    limit  = true;
  }

  memcpy(C_str, Pascal_str, length);
  C_str[length] = '\0';

  if (Is_limited_length) *Is_limited_length = limit;

  return C_str;
}

/*------------------------------------------------------------------------------------------------*/

/*
    Pascal_str       : Beginning address of Pascal style string area.
    Pascal_max_size  : Size of Pascal style string area, included the string length.
    C_str            : Beginning address of C style string area.
    Is_limited_length: This a pointer wherewith the function indicates that limitation occurred.
*/

FUNC(size_t) gp_Pstr_from_str(uint8_t *Pascal_str, size_t Pascal_max_size, const char *C_str, bool *Is_limited_length) {
  size_t     length;
  bool limit=false;

  assert(Pascal_str);
  assert(C_str);

  length = strlen(C_str);

  if (length >= Pascal_max_size) {
    /* Truncate the length. */
    length = Pascal_max_size - 1;
    limit  = true;
  }

  /* Store the string length. */
  *Pascal_str++ = length;	/* The real beginning of string. */
  memcpy(Pascal_str, C_str, length);

  if (Is_limited_length) *Is_limited_length = limit;

  return length;
}

/*------------------------------------------------------------------------------------------------*/

/*
    C_str            : Beginning address of C style string area.
    Pascal_max_size  : Size of Pascal style string area, included the string length.
    Is_limited_length: This a pointer wherewith the function indicates that limitation occurred.
*/

FUNC(size_t) gp_strlen_Plimit(const char *C_str, size_t Pascal_max_size, bool *Is_limited_length) {
  size_t     length;
  bool limit=false;

  assert(C_str);

  length = strlen(C_str);

  if (length >= Pascal_max_size) {
    length = Pascal_max_size - 1;
    limit  = true;
  }

  if (Is_limited_length) *Is_limited_length = limit;

  return length;
}

/*------------------------------------------------------------------------------------------------*/

FUNC(size_t) gp_align_text(char *Buffer, size_t Buffer_length, size_t Current_length, size_t Aligned_to_length) {
  int    lp;
  size_t length;

  if ((Current_length < (Buffer_length - 1)) && (Aligned_to_length > Current_length)) {
    Buffer_length -= Current_length;
    length = Aligned_to_length - Current_length;

    if (length >= Buffer_length) {
      length = Buffer_length - 1;
    }

    lp = snprintf(&Buffer[Current_length], Buffer_length - Current_length, "%*s", (int)length, " ");
    return lp > 0 ? Current_length + lp : Current_length;
  }

  return Current_length;
}

/*------------------------------------------------------------------------------------------------*/

CFUNC(size_t) gp_exclamation(char *Buffer, size_t Buffer_length, size_t Current_length, const char *Format, ...) {
  int     lp;
  size_t  length;
  va_list ap;

  length = gp_align_text(Buffer, Buffer_length, Current_length, EXPLANATION_DISTANCE);
  va_start(ap, Format);
  lp = vsnprintf(&Buffer[length], Buffer_length - length, Format, ap);
  va_end(ap);
  return ((lp >= 0) ? (length + lp) : Current_length);
}

/*------------------------------------------------------------------------------------------------*/

/* fetch the absolute path of the filename */

FUNC(char*) gp_absolute_path(char *File_name) {
#ifdef HAVE_WINDOWS_H

  /* It would be better to test for GetFullPathName, but the test won't
     work with a cross compiler. So if windows.h exists, we assume that
     GetFullPathName is available. */ 

  #define FILE_BUFFER_SIZE 512
  char  file_buffer[FILE_BUFFER_SIZE];
  char *file_ptr;
  int   num_chars;

  num_chars = GetFullPathName(File_name, FILE_BUFFER_SIZE, file_buffer, &file_ptr);
  if (!num_chars) {
    gp_error("Can't fetch full path of %s.", File_name);
    return File_name;
  }
  else {
    return GP_Strdup(file_buffer);
  }
#else
  #ifdef HAVE_REALPATH
  char *resolved_name;

  resolved_name = realpath(File_name, NULL);
  if (resolved_name == NULL) {
    gp_error("Can't fetch full path of \"%s\".", File_name);
    return File_name;
  }
  else {
    return resolved_name;
  }

  #else

  if (absolute_path_warning) {
    gp_warning("Host system does not support absolute paths.");
    absolute_path_warning = false;
  }
  
  return File_name;

  #endif
#endif
}

/*------------------------------------------------------------------------------------------------*/

FUNC(void) gp_exit_if_arg_an_option(const struct option *Options, int Opt_max_index, int Opt_index,
                         const char *Opt_string, int Opt_char, const char *Command) {
  const struct option *opt;
  int                  i;
  char                 cmd[8];

  if (Opt_index < 0) {
    /* This is likely a short option. */
    opt = Options;
    i = 0;

    while (opt->name) {
      if (opt->val == Opt_char) {
        Opt_index = i;
        cmd[0] = '-';
        cmd[1] = (char)Opt_char;
        cmd[2] = '\0';
        Command = cmd;
        break;
      }

      ++opt;
      ++i;
    }
  }

  if (Opt_index < 0) {
    /* This not an option. */
    return;
  }

  if (Opt_index >= Opt_max_index) {
    fprintf(stderr, "%s:%u -- Fatal error: opt_index == %i (Only valid if opt_index < %i.)\n",
            __FILE__, __LINE__, Opt_index, Opt_max_index);
    exit(1);
  }

  if (Options[Opt_index].has_arg == no_argument) {
    return;
  }

  if ((Opt_string == NULL) || (Opt_string[0] != '-') || (Opt_string[1] == '\0')) {
    /* This opt_string can not be option. */
    return;
  }

  /* opt_string == "-." */
  opt = Options;
  while (opt->name) {
    if ((Opt_string[1] == '-') && (Opt_string[2] != '\0')) {
      /* Opt_string == "--." This is a long option? */
      if (strcmp(&Opt_string[2], opt->name) == 0) {
        fprintf(stderr, "Error: This option may not be parameter of the \"%s\" option: \"--%s\" (\"%s\")\n",
                Command, opt->name, Opt_string);
        exit(1);
      }
    }

    if (isalnum(opt->val)) {
      /* Opt_string == "-X" This is a short option? */
      if (Opt_string[1] == (char)opt->val) {
        fprintf(stderr, "Error: This option may not be parameter of the \"%s\" option: \"-%c\" (\"%s\")\n",
                Command, (char)opt->val, Opt_string);
        exit(1);
      }
    }

    ++opt;
  }
}
